//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// Logger program for the SY2527 power supply
//
// Hardware setup:
// - Power:                     ON, LOCAL
// - "Local" switch:            LOCAL ENABLE
// - "Interlock" switch:        OFF
// - HV Board:                  50 Ohm dummy load on HV-ENABLE
// -------------------------------------------------------------------
// $Id: caenlogger.cxx,v 1.5 2004/12/16 11:17:29 loreti Exp $
// -------------------------------------------------------------------

#include <fstream>
#include <iomanip>
#include <cctype>
#include <cstdlib>
#include "sy2527.hh"

extern "C" {
  #include <unistd.h>
  #include "inter.h"
}

using namespace std;

namespace {
  const string  name    = "Test";
  const string  address = IPADDR;
  unsigned long boardNr = 1;
  int           maxChan = 12;
}

static void printBdStatus(
  ofstream &   outf,
  unsigned int status
) {
  // Prints on the standard output the meaning of the bits of the
  // board status

  if (status & 0x01) outf << "FAIL"; else outf << "OK";
  if (status & 0x02) outf << ", CKSUMERROR";
  if (status & 0x04) outf << ", HVCALERROR";
  if (status & 0x08) outf << ", TEMPCALERR";
  if (status & 0x10) outf << ", UNDERTEMP";
  if (status & 0x20) outf << ", OVERTEMP";
  outf << endl;
}

static void monitorBoard(
  ofstream &      outf,
  const caenPST & ps,
  int             board
) {
  // Gets and prints board temperature and status

  double temp         = ps.getTemp(board);
  unsigned int status = ps.getBdStatus(board);

  outf << "Board " << board << ", T=" << temp << " deg., ";
  printBdStatus(outf, status);
}

static void monitorChannel(
  ofstream &      outf,
  const caenPST & ps,
  int             board,
  int             channel
) {
  // Gets and prints relevant values and status

  double v            = ps.getV(board, channel);
  double i            = ps.getI(board, channel);
  long   j            = static_cast<long>( i * 1000.0 + 0.5 );
  unsigned int status = ps.getChStatus(board, channel);

  outf << setw(5) << v << ' ';
  outf << setw(5) << j << ' ';
  outf << hex << setw(3) << status << dec << ' ';
 }

int main(
  int argc,
  char * argv[]
) {
  if (argc != 3) {
    cout << "Usage: " << argv[0] << " <filename> <delay (s)>\n";
    return 0;
  }

  ofstream outf(argv[1]);
  if (!outf) {
    cout << "Couldn't open file \"" << argv[1] << "\"\n";
    return 0;
  }

  unsigned long int delay = strtoul(argv[2], 0, 0);

  inter();                      // Establish the CTRL-C handler

#if __GNUG__ < 3
  outf << setprecision(1);
#else
  outf << fixed << setprecision(1);
#endif

  // - Opens the connection toward the power supply;
  // - Prints, every 5 seconds, the current settings for board
  //   "boardNr" and board  "boardNr" channel "channNr".

  while (true) {
    try {
      caenPST ps(name, address);

      for (int chan = 0;   chan < maxChan;   ++chan) {
        monitorChannel(outf, ps, boardNr, chan);
      }

      outf << sy2527::timeStamp() << "- ";
      monitorBoard(outf, ps, boardNr);
    } catch (const exception & e) {
      cerr << "*** EXCEPTION:" << e.what() << endl;
      return 0;
    } // try/catch

    sleep(delay);
  } // forever loop
}
