//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// Exerciser program for the SY2527 power supply
//
// Hardware setup:
// - Power:                     ON, LOCAL
// - "Local" switch:            LOCAL ENABLE
// - "Interlock" switch:        OFF
// - HV Board:                  50 Ohm dummy load on HV-ENABLE
// -------------------------------------------------------------------
// $Id: exerciser.cxx,v 1.11 2004/12/16 11:17:29 loreti Exp $
// -------------------------------------------------------------------

#include <cctype>
#include <cstdlib>
#include "sy2527.hh"
#include "keyword.hh"

extern "C" {
  #include <unistd.h>
}

using std::cerr;
using std::cin;
using std::copy;
using std::cout;
using std::endl;
using std::exception;
using std::exit;
using std::getline;
using std::isspace;
using std::ostream_iterator;
using std::string;
using std::strtoul;

namespace {
  const string  defaultName    = "Test";
  const string  defaultAddress = IPADDR;
  unsigned long boardNr        = 1;
  unsigned long channNr        = 0;
  char          spaceLike[]    = " \t\f\r\v"; // "C" locale definition
}

static void printBdStatus(
  unsigned int status
) {
  // Prints on the standard output the meaning of the bits of the
  // board status

  if (status & 0x01) cout << "FAIL"; else cout << "OK";
  if (status & 0x02) cout << ", CKSUMERROR";
  if (status & 0x04) cout << ", HVCALERROR";
  if (status & 0x08) cout << ", TEMPCALERR";
  if (status & 0x10) cout << ", UNDERTEMP";
  if (status & 0x20) cout << ", OVERTEMP";
  cout << endl;
}

static bool monitorBoard(
  const caenPST & ps,
  int             board
) {
  // Gets and prints board temperature and status; returns "true" if
  // all is OK (status == 0)

  double temp         = ps.getTemp(board);
  unsigned int status = ps.getBdStatus(board);

  cout << "Board " << board << ": Temp. " << temp
       << " degrees; status ";
  printBdStatus(status);
  return status == 0;
}

static void printChStatus(
  unsigned int status
) {
  // Prints on the standard output the meaning of the bits of the
  // channel status

  if (status & 0x001) cout << "ON"; else cout << "OFF";
  if (status & 0x002) cout << ", RAMPUP";
  if (status & 0x004) cout << ", RAMPDOWN";
  if (status & 0x008) cout << ", OVERCURR";
  if (status & 0x010) cout << ", OVERVOLT";
  if (status & 0x020) cout << ", UNDERVOLT";
  if (status & 0x040) cout << ", EXTERNTRIP";
  if (status & 0x080) cout << ", MAXHV";
  if (status & 0x100) cout << ", EXTERNDISABLE";
  if (status & 0x200) cout << ", INTERNTRIP";
  if (status & 0x400) cout << ", CALERROR";
  if (status & 0x800) cout << ", UNPLUGGED";
  cout << endl;
}

static int monitorChannel(
  const caenPST & ps,
  int             board,
  int             channel
) {
  // Gets and prints relevant values and status; returns -1 in case of
  // error, 0 if all is OK and the channel is OFF, 1 if all is OK and
  // the channel is ON.

  double v            = ps.getV(board, channel);
  double i            = ps.getI(board, channel);
  double v0           = ps.getV0(board, channel);
  double i0           = ps.getI0(board, channel);
  double trip         = ps.getTrip(board, channel);
  double rup          = ps.getRup(board, channel);
  double rdn          = ps.getRdn(board, channel);
  double svmax        = ps.getSVMax(board, channel);
  unsigned int status = ps.getChStatus(board, channel);

  cout << "Bd " << board << ", ch " << channel
       << ": V=" << v << ", I=" << i << ", V0="
       << v0 << ", I0=" << i0 << ", Trip=" << trip
       << ", RUp=" << rup << ", Rdn=" << rdn
       << ", SVMax=" << svmax << "\nChannel status: ";
  printChStatus(status);
  if (status < 2) return status; else return -1;
}

static void set(
  const caenPST &  ps,
  int              board,
  int              channel,
  const KeyTable & attribKT,
  const string     attrib[],
  size_t           nAttribs,
  string &         continuation
) {
  string            answer;
  string::size_type l;
  int               i;
  double            value = 0.0;

  do {
    if (continuation.empty()) {
      cout << "Set what ? ";
      if (!getline(cin, answer)) exit(0);
    } else {
      answer = continuation;
      continuation.erase();
    }
  } while (answer.empty()  ||
           (l = answer.find_first_not_of(spaceLike)) == string::npos);

  // Strips leading and trailing whitespace

  if (l) answer.erase(0, l);
  if ((l = answer.find_first_of(spaceLike)) != string::npos) {
    continuation = string(answer, l);
    answer.erase(l);
  } else {
    continuation = "";
  }

  switch (attribKT.getId(answer, &i)) {
    case KeyTable::KT_OK:
      char * pc;

      if (i < 8) {
        while (true) {
          if (continuation.empty()) {
            cout << "Value ? ";
            if (!getline(cin, answer)) exit(0);
            if (answer.empty()) continue;
          } else {
            answer = continuation;
            continuation.erase();
          }
          if (answer.find_first_not_of(spaceLike) == string::npos) continue;
          value = std::strtod(answer.c_str(), &pc);
          if (isspace(static_cast<unsigned char>(*pc)) ||
              *pc == '\0') break;
          else cout << "Invalid: \"" << answer << "\"" << endl;
        }
      }
      switch (i) {
        case 0:       // V0
          ps.setV0(board, channel, value);
          break;
        case 1:       // I0
          ps.setI0(board, channel, value);
          break;
        case 2:       // V1
          ps.setV1(board, channel, value);
          break;
        case 3:       // I1
          ps.setI1(board, channel, value);
          break;
        case 4:       // Trip
          ps.setTrip(board, channel, value);
          break;
        case 5:       // Ramp Up
          ps.setRup(board, channel, value);
          break;
        case 6:       // Ramp Down
          ps.setRdn(board, channel, value);
          break;
        case 7:       // Software VMax
          ps.setSVMax(board, channel, value);
          break;
        case 8:       // Channel On
          ps.setOn(board, channel);
          break;
        case 9:       // Channel Off
          ps.setOff(board, channel);
          break;
        default:
          cout << "Can't happen - file " << __FILE__
               << ", line " << __LINE__ << endl;
          break;
      }
      break;
    case KeyTable::KT_UNDEF:
      cout << "Undefined keyword";
      goto list;
    case KeyTable::KT_AMBIG:
      cout << "Ambiguous keyword";
  list:
      cout << "; allowed are: " << endl;
      copy(attrib, attrib + nAttribs,
           ostream_iterator<string>(cout, " "));
      cout << endl;
      break;
    default:
      cout << "Can't happen - file " << __FILE__
           << ", line " << __LINE__ << endl;
      break;
  }
}

static void ask(
  const caenPST & ps,
  int             board,
  int             channel
) {
  const string action[] = { "Read", "Set", "Exit", "Quit" };
  const size_t nActions = sizeof(action) / sizeof(action[0]);
  const string attrib[] = { "V0", "I0", "V1", "I1", "Trip",
                            "Rup", "Rdn", "SVMax", "On", "Off" };
  const size_t nAttribs = sizeof(attrib) /sizeof(attrib[0]);

  KeyTable actionKT, attribKT;

  try {
    actionKT = KeyTable(action, 0, nActions);
    attribKT = KeyTable(attrib, 0, nAttribs);
  } catch (const exception & e) {
    cerr << "EXCEPTION from KeyTable::KeyTable() - "
         << e.what() << endl;
    return;
  }

  while (true) {
    string            answer;
    string            continuation;
    string::size_type l;
    int               i;

    cout << "Action ? ";
    if (!getline(cin, answer)) exit(0);
    if (answer.empty()) continue;

    // Strips leading and trailing whitespace

    if ((l = answer.find_first_not_of(spaceLike)) == string::npos) continue;
    if (l) answer.erase(0, l);
    if ((l = answer.find_first_of(spaceLike)) != string::npos) {
      continuation = string(answer, l);
      answer.erase(l);
    } else {
      continuation = "";
    }

    switch (actionKT.getId(answer, &i)) {
      case KeyTable::KT_OK:
        switch (i) {
          case 0:               // Read
            monitorBoard(ps, board);
            monitorChannel(ps, board, channel);
            break;
          case 1:               // Set
            set(ps, board, channel, attribKT, attrib, nAttribs, continuation);
            break;
          case 2:               // Exit
          case 3:               // Quit
            cout << "Bye\n";
            return;
          default:
            cout << "Can't happen - file " << __FILE__
                 << ", line " << __LINE__ << endl;
            break;
        }
        break;
      case KeyTable::KT_UNDEF:
        cout << "Undefined keyword";
        goto list;
      case KeyTable::KT_AMBIG:
        cout << "Ambiguous keyword";
    list:
        cout << "; allowed are: ";
        copy(action, action + nActions,
             ostream_iterator<string>(cout, " "));
        cout << endl;
        break;
      default:
        cout << "Can't happen - file " << __FILE__
             << ", line " << __LINE__ << endl;
        break;
    }
  }
}

int main() {
  string name, address, answer;

  cout << "Enter name (default \"" << defaultName << "\"): ";
  if (!getline(cin, name)) return 0;
  if (name.empty()) name = defaultName;

  cout << "Enter address (default \"" << defaultAddress << "\"): ";
  if (!getline(cin, address)) return 0;
  if (address.empty()) address = defaultAddress;

  while(true) {
    cout << "Enter board number (default " << boardNr << "): ";
    if (!getline(cin, answer)) return 0;
    if (answer.empty()) break;

    char *        pc;
    unsigned long b = strtoul(answer.c_str(), &pc, 0);

    if (!isspace(static_cast<unsigned char>(*pc)) && *pc != '\0') {
      cerr << "Invalid: \"" << answer << "\"\n";
    } else {
      boardNr = b;
      break;
    }
  }

  while(true) {
    cout << "Enter channel number (default " << channNr << "): ";
    if (!getline(cin, answer)) return 0;
    if (answer.empty()) break;

    char *        pc;
    unsigned long c = strtoul(answer.c_str(), &pc, 0);

    if (!isspace(static_cast<unsigned char>(*pc)) && *pc != '\0') {
      cerr << "Invalid: \"" << answer << "\"\n";
    } else {
      channNr = c;
      break;
    }
  }
  cout << '\n';
  sy2527::version();
  cout << endl;

  try {
    caenPST ps(name, address);

    if (!ps.testBC(boardNr, channNr)) {
      cerr << "\nBoard " << boardNr << " channel " << channNr
           << ": invalid\n";
      return 0;
    }

#ifndef NDEBUG
    cout << endl;
#endif
    monitorBoard(ps, boardNr);
    monitorChannel(ps, boardNr, channNr);
    ask(ps, boardNr, channNr);
  } catch (const exception & e) {
    cerr << "*** EXCEPTION:" << e.what() << endl;
  }
}
