//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// -------------------------------------------------------------------
// $Id: keyword.hh,v 1.2 2005/01/21 07:31:45 loreti Exp $
// -------------------------------------------------------------------

#ifndef KEYWORD_HH_
#define KEYWORD_HH_

#include <algorithm>
#include <functional>
#include <iterator>
#include <stdexcept>
#include <string>
#include <utility>
#include <vector>

#include <cctype>

using std::back_inserter;
using std::runtime_error;
using std::string;
using std::transform;
using std::vector;

// A class for the keyword entries.
// An entry is abstracted as:
// - an alphanumeric keyword;
// - a minimum number of characters to be matched in that identifier
//   in order to be uniquely identified in the current context;
// - an integer value to be returned in case of match.

class Key {
  friend class KeyTable;

private:
  string key;                   // Keyword
  string::size_type match;      // Number of characters to match
  int id;                       // User-defined data

public:
  Key(const string & s, const int i) : key(s), match(0), id(i) {}
  ~Key() {}

  const string &          getKey()   const { return key;   }
        string::size_type getMatch() const { return match; }
        int               getId()    const { return id;    }

  bool operator <  (const Key & rhs) const { return key <  rhs.key; }
  bool operator == (const Key & rhs) const { return key == rhs.key; }
};

// A context for keyword search.
// A keyword table is a (sorted) vector of Keys, where the minimum
// number of characters needed to identify each keyword has been
// computed; if the constructor fails, an exception is thrown.
// A method getId() is supplied to search the table for a given
// string.

class KeyTable {
public:
  enum eFlags { KT_OK = 0, KT_UNDEF, KT_AMBIG, KT_DUPL, KT_INVALID };

private:
  static const string invInp;
  static const string dupKey;
  static const string ambKey;

  vector<Key> keyVec;
  enum eFlags buildTable();

public:
  KeyTable(const string * = 0, const int * = 0, const size_t = 0);
  ~KeyTable() {}
  enum eFlags getId(const string &, int *) const;
};

const string KeyTable::invInp("invalid input");
const string KeyTable::dupKey("duplicate keyword");
const string KeyTable::ambKey("ambiguous keyword");

// A function object used to translate characters to lowercase

struct low : public std::unary_function<unsigned char, unsigned char> {
  unsigned char operator () (unsigned char c) {
    return std::tolower(static_cast<unsigned char>(c)); }
};

// *** Methods ***

// The KeyTable constructor.  Fills the set of internal keys with the
// supplied strings converted to lowercase, and the corresponding
// numerical values supplied by the user (or, if "ids" is NULL, with
// the sequential number of the key inside the string array); then
// calls buildTable() to compute the minimum number of characters
// needed in order to uniquely identify every keyword.

KeyTable::KeyTable(
  const string   keys[],        // Array of keywords
  const int    * ids,           // Array of integer values, or NULL
  const size_t   num            // Dimension of the above arrays
) {
  if (num > 0) {
    keyVec.reserve(num);

    for (size_t i = 0;   i < num;   i++) {
      int id = (ids ? ids[i] : i);

      string lStr;
      transform((keys[i]).begin(), (keys[i]).end(),
                back_inserter(lStr), low());

      keyVec.push_back( Key(lStr, id) );
    }

    std::sort(keyVec.begin(), keyVec.end());
    vector<Key>::iterator finis = keyVec.end();
    if (std::unique(keyVec.begin(), finis) != finis) {
      throw runtime_error(dupKey);
    }

    switch (buildTable()) {
      case KT_INVALID: throw runtime_error(invInp);
      case KT_AMBIG:   throw runtime_error(ambKey);
      default:         break;
    }
  }
}

enum KeyTable::eFlags KeyTable::buildTable()
{
  if (keyVec.empty()) return KT_INVALID;

  vector<Key>::size_type prev, curr;
  keyVec[0].match = 1;

  for (prev = 0,  curr = 1;   curr != keyVec.size();   prev = curr++) {
    const char *p0, *p1, *p2;

    p1 =      keyVec[prev].key.c_str();
    p0 = p2 = keyVec[curr].key.c_str();
    while (*p1 == *p2) {        // Cannot find a duplicate (this
      p1++;                     //   has been tested in the ctor)
      p2++;
    }
    if (*p1 == '\0') return KT_AMBIG;

    keyVec[curr].match = ++p2 - p0;
    if (keyVec[curr].match > keyVec[prev].match) {
      keyVec[prev].match = keyVec[curr].match;
    }
  }

  return KT_OK;
}

// getId() compares the given string (converted to lowercase) with the
// known keywords.  Returns KT_OK if a match is found, storing the
// corresponding integer value into "val"; if an error occurs, "val"
// is set to zero and the function value is either KT_UNDEF (no match)
// or KT_AMBIG (more than one match).

enum KeyTable::eFlags KeyTable::getId(
  const string & key,
  int          * val
) const {
  *val = 0;

  string lStr;
  transform(key.begin(), key.end(), back_inserter(lStr), low());

  vector<Key>::const_iterator it, finis = keyVec.end();
  for (it = keyVec.begin();   it != finis;   it++) {
    if (lStr.compare( (it->key).substr(0, lStr.size()) ) == 0) {
      if (lStr.size() < it->match) return KT_AMBIG;
      *val = it->id;
      return KT_OK;
    }
  }
  return KT_UNDEF;
}

#endif // KEYWORD_HH_
