//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// The code for the procedures defined in sy2527.hh
// -------------------------------------------------------------------
// $Id: sy2527.cxx,v 1.10 2004/02/25 09:56:28 loreti Exp $
// -------------------------------------------------------------------

#include <algorithm>
#include <functional>
#include <stdexcept>
#include <cctype>
#include <cstdlib>
#include "sy2527.hh"

using std::cerr;
using std::cout;
using std::endl;
using std::free;
using std::ostream;
using std::string;

namespace sy2527 {
  static const char versionNR[] = "1.0";
  static const char WLIB[]      = "2.11";
  static const char CLIB[]      = "1.6";

  void version(
    std::ostream & o            // Defaults to std::cout
  ) {

    // Printout on an output stream of the software version numbers

    o << "- CAEN HV Wrapper library version " << WLIB
      << "\n- HS CAENet library version " << CLIB
      << "\n- SY2527 classes version " << versionNR
      << ", compiled " << __DATE__ << endl;
  }
} // namespace sy2527

/*--------------------------------*
 | Methods for the class "caenBD" |
 *--------------------------------*/

void caenBD::out(
  ostream & o                   // Defaults to std::cout
) const {
  // Prints on the output stream "o" (default: the standard output)
  // the internal values; public method.

  o << "type " << _type << ',' << _desc << "; " << _nChan
    << " channels, serial nr. " << _serial;
}

ostream & operator << (
  ostream &      o,
  const caenBD & b
) {
  // Prints on the output stream "o" the internal values; output
  // operator.

  b.out(o);
  return o;
}

/*------------------------------------------------------*
 | Static constants and methods for the class "caenPST" |
 *------------------------------------------------------*/

const char caenPST::_username[] = "admin";
const char caenPST::_password[] = "admin";

// The constructor

caenPST::caenPST(
  const string & n,
  const string & a
) : _systemName(n), _ipAddress(a), _nBoards(0) {
  CAENHVRESULT err;
  err = CAENHVInitSystem(n.c_str(), LINKTYPE_TCPIP,
                         const_cast<char *>(a.c_str()),
                         _username, _password);
  if (err) {
    throw std::runtime_error( CAENHVGetError(n.c_str()) );
#ifdef CAENDEBUG
  } else {
    cout << sy2527::timeStamp() << "- SY2527 connected to " << a
         << endl;
#endif // CAENDEBUG
  }

  unsigned short   nSlots;
  unsigned short * nChan;
  unsigned short * serial;
  char *           mode;
  char *           desc;
  unsigned char *  minRel;
  unsigned char *  maxRel;

  err = CAENHVGetCrateMap(n.c_str(), &nSlots, &nChan, &mode,
                          &desc, &serial, &minRel, &maxRel);
  if (err) {
    cerr << "GetCrateMap failed: " << CAENHVGetError(n.c_str())
         << '\n';
  } else {
    char *saveMode = mode;
    char *saveDesc = desc;

    _nBoards = nSlots;
    _boards.reserve(_nBoards);

    for (unsigned int i = 0;   i < _nBoards;   ++i) {
      caenBD * pBD;

      pBD = *mode ? new caenBD(nChan[i], serial[i], mode, desc) : 0;
      _boards.push_back(pBD);

#ifndef NDEBUG
      cout << "Board " << i << ": ";
      if (*mode) cout << *pBD; else cout << "missing";
      cout << endl;
#endif

      while (*mode++) {}
      while (*desc++) {}
    }

    free(nChan);
    free(serial);
    free(saveMode);
    free(saveDesc);
    free(minRel);
    free(maxRel);
  }
}

// An helper function object, to delete instances of caenBD ADT's

struct deletePointed : public std::unary_function<caenBD *, void> {
  void operator() (caenBD * p) { delete p; }
};

// The destructor

caenPST::~caenPST() {
  CAENHVRESULT err;

  err = CAENHVDeinitSystem(_systemName.c_str());
  if (err) cerr << CAENHVGetError(_systemName.c_str()) << '\n';
  std::for_each(_boards.begin(), _boards.end(), deletePointed());

#ifdef CAENDEBUG
  cout << sy2527::timeStamp() << "- SY2527 disconnected from "
       << _ipAddress << endl;
#endif // CAENDEBUG
}
