//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// Header file for the CAEN SY2527 power supply communication handler
// -------------------------------------------------------------------
// $Id: sy2527.hh,v 1.14 2004/02/25 09:56:28 loreti Exp $
// -------------------------------------------------------------------

#ifndef SY2527_HH_
#define SY2527_HH_

#include <iostream>
#include <string>
#include <vector>
#include <ctime>

extern "C" {
  #include "CAENHVWrapper.h"
}

namespace sy2527 {
  // "INVALID" is a value that cannot be returned from a CAEN library
  // procedure

  const CAENHVRESULT INVALID = -1;

  void version (std::ostream & = std::cout);
  inline std::string timeStamp();

  inline std::string timeStamp() { // Returns a timestamp as a std::string
    time_t now = std::time(0);
    char * pc = std::asctime( std::localtime( &now ));
    return std::string(pc+4, pc+20);
  }
} // namespace sy2527

// CAEN high voltage board
// - Data: type, description, serial number, number of HV channels.
// - Methods: constructor and destructor; printout to an output
//   stream; an accessor to get the number of HV channels.
// - Note: the dtor is virtual provided that specialized board classes
//   may inherit from caenBD (not done here since we have only the
//   A1511 HV board).

class caenBD {
private:
  const std::string  _type;
  const std::string  _desc;
  const unsigned int _serial;
  const unsigned int _nChan;

public:
  caenBD(unsigned short n, unsigned short s,
         const char * m, const char * d)
    : _type(m), _desc(d), _serial(s), _nChan(n) {}
  virtual ~caenBD() {}

  void out(std::ostream & = std::cout) const;

  unsigned int getNChan() const { return _nChan; }
};

// CAEN power supply controlled over TCP/IP
// - Data: username and password needed in order to access the crate
//   (static); symbolic name; IP address; a vector of caenBD
//   instances, one for every board; and the maximum number of boards
//   in the crate.
// - Methods: constructor and destructor; procedures to get and set
//   various values related to a specific board, or to a specific
//   channel of a given board.
// - Private methods: procedures internally used by the public
//   methods.
// - Notes: virtual destructor; copy constructor and assignment
//   operator declared private and never defined in order to prevent
//   their use.

class caenPST {
private:
  static const char     _username[];
  static const char     _password[];
  const  std::string    _systemName;
  const  std::string    _ipAddress;
  std::vector<caenBD *> _boards;
  unsigned int          _nBoards;

public:
  caenPST(const std::string &, const std::string &);
  virtual ~caenPST();

  inline double       getTemp(unsigned short)     const;
  inline unsigned int getBdStatus(unsigned short) const;

  inline double       getV(unsigned int, unsigned short)        const;
  inline double       getI(unsigned int, unsigned short)        const;
  inline double       getV0(unsigned int, unsigned short)       const;
  inline double       getI0(unsigned int, unsigned short)       const;
  inline double       getV1(unsigned int, unsigned short)       const;
  inline double       getI1(unsigned int, unsigned short)       const;
  inline double       getTrip(unsigned int, unsigned short)     const;
  inline double       getRup(unsigned int, unsigned short)      const;
  inline double       getRdn(unsigned int, unsigned short)      const;
  inline double       getSVMax(unsigned int, unsigned short)    const;
  inline unsigned int getChStatus(unsigned int, unsigned short) const;

  inline bool setV0(unsigned int, unsigned short, float)    const;
  inline bool setI0(unsigned int, unsigned short, float)    const;
  inline bool setV1(unsigned int, unsigned short, float)    const;
  inline bool setI1(unsigned int, unsigned short, float)    const;
  inline bool setTrip(unsigned int, unsigned short, float)  const;
  inline bool setRup(unsigned int, unsigned short, float)   const;
  inline bool setRdn(unsigned int, unsigned short, float)   const;
  inline bool setSVMax(unsigned int, unsigned short, float) const;
  inline bool setOnOff(unsigned int, unsigned short, bool)  const;
  inline bool setOn(unsigned int, unsigned short)           const;
  inline bool setOff(unsigned int, unsigned short)          const;

  unsigned int getNBoards() const { return _nBoards; }

  // Inline test for the existence of a given board.  Returns the
  // address of the related caenBD instance, or 0.
  caenBD * testB(unsigned int b) const {
    return b < _nBoards ? _boards[b] : 0; }

  // Inline test for the existence of a given channel in a given
  // board.
  bool testBC(unsigned int b, unsigned short c) const {
    caenBD * pBD = testB(b);
    return pBD != 0  &&   c < pBD->getNChan(); }

private:
  template <typename T>
  T getCVal(unsigned int, unsigned short, const char *, const T &) const;

  template <typename T>
  T getBVal(unsigned short, const char *, const T &) const;

  template <typename T>
  bool setCVal(unsigned int, unsigned short, const char *, const T &) const;

  caenPST(const caenPST &);
  caenPST & operator = (const caenPST &);
};

// A private procedure that gets the value of the parameter "name" for
// the board "board"; "T" is the ADT returned from the CAEN library
// interface.  If an error is detected (read error, a non-existent
// board, ...) a warning is printed on the standard error stream and
// "errVal" is returned.

template <typename T> T caenPST::getBVal(
  unsigned short board,
  const char *   name,
  const T &      errVal
) const {
  if (testB(board)) {
    CAENHVRESULT err;
    T result;

    err = CAENHVGetBdParam(_systemName.c_str(), 1, &board,
                           name, &result);
    if (err != CAENHV_OK) {
#ifndef NDEBUG
      std::cerr << "getBVal(bd=" << board << ", op=\"" << name
                << "\") failed: "
                << CAENHVGetError(_systemName.c_str()) << '\n';
#endif // NDEBUG
    } else {
#ifdef CAENDEBUG
      std::cout << sy2527::timeStamp() << "- SY2527 getBVal(bd="
                << board << ", op=\"" << name << "\") = " << result
                << std::endl;
#endif // CAENDEBUG

      return result;
    }
  }
  return errVal;
}

// A private procedure that gets the value of the parameter "name" for
// the channel "channel" of the board "board"; "T" is the ADT returned
// from the CAEN library interface.  If an error is detected (read
// error, a non-existent board, ...) a warning is printed on the
// standard error stream and "errVal" is returned.

template <typename T> T caenPST::getCVal(
  unsigned int   board,
  unsigned short channel,
  const char *   name,
  const T &      errVal
) const {
  caenBD * pBD = testB(board);

  if (pBD   &&   channel < pBD->getNChan()) {
    CAENHVRESULT err;
    T result;

    err = CAENHVGetChParam(_systemName.c_str(), board, name,
                           1, &channel, &result);
    if (err != CAENHV_OK) {
#ifndef NDEBUG
      std::cerr << "getCVal(bd=" << board << ", ch=" << channel
                << ", op=\"" << name << "\") failed: "
                << CAENHVGetError(_systemName.c_str()) << '\n';
#endif // NDEBUG
    } else {
#ifdef CAENDEBUG
      std::cout << sy2527::timeStamp() << "- SY2527 getCVal(bd=" << board
                << ", ch=" << channel << ", op=\"" << name << "\") = "
                << result << std::endl;
#endif // CAENDEBUG

      return result;
    }
  }
  return errVal;
}

// A private procedure that sets to "value" the value of the parameter
// "name" for the channel "channel" of the board "board"; "T" is the
// ADT passed as a formal parameter to the CAEN library interface.
// "false" is returned if any error is detected, "true" otherwise.

template <typename T> bool caenPST::setCVal(
  unsigned int   board,
  unsigned short channel,
  const char *   name,
  const T &      value
) const {
#ifdef CAENDEBUG
      std::cout << sy2527::timeStamp() << "- SY2527 setCVal(bd=" << board
                << ", ch=" << channel << ", op=\"" << name
                << "\", val=" << value << ")" << std::endl;
#endif // CAENDEBUG

  caenBD * pBD     = testB(board);
  CAENHVRESULT err = sy2527::INVALID;

  if (pBD   &&   channel < pBD->getNChan()) {
    err = CAENHVSetChParam(_systemName.c_str(), board, name, 1,
                           &channel, const_cast<T *>(&value));
#ifndef NDEBUG
    if (err != CAENHV_OK) {
      std::cerr << "setCVal(bd=" << board << ", ch=" << channel
                << ", op=\"" << name << "\", val=" << value
                << ") failed: "
                << CAENHVGetError(_systemName.c_str()) << '\n';
    }
#endif // NDEBUG
  }
  return err == CAENHV_OK;
}

// Getter methods

inline double caenPST::getTemp(
  unsigned short board
) const {
  return getBVal<float>(board, "Temp", 0.0);
}

inline unsigned int caenPST::getBdStatus(
  unsigned short board
) const {
  return getBVal<unsigned int>(board, "BdStatus", 0);
}

inline unsigned int caenPST::getChStatus(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<unsigned int>(board, ch, "Status", 0);
}

inline double caenPST::getV(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<float>(board, ch, "VMon", 0.0);
}

inline double caenPST::getI(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<float>(board, ch, "IMon", 0.0);
}

inline double caenPST::getV0(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<float>(board, ch, "V0Set", 0.0);
}

inline double caenPST::getI0(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<float>(board, ch, "I0Set", 0.0);
}

inline double caenPST::getV1(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<float>(board, ch, "V1Set", 0.0);
}

inline double caenPST::getI1(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<float>(board, ch, "I1Set", 0.0);
}

inline double caenPST::getTrip(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<float>(board, ch, "Trip", 0.0);
}

inline double caenPST::getRup(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<float>(board, ch, "RUp", 0.0);
}

inline double caenPST::getRdn(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<float>(board, ch, "RDWn", 0.0);
}

inline double caenPST::getSVMax(
  unsigned int   board,
  unsigned short ch
) const {
  return getCVal<float>(board, ch, "SVMax", 0.0);
}

// Setters methods

inline bool caenPST::setV0(
  unsigned int   board,
  unsigned short ch,
  float          val
) const {
  return setCVal<float>(board, ch, "V0Set", val);
}

inline bool caenPST::setI0(
  unsigned int   board,
  unsigned short ch,
  float          val
) const {
  return setCVal<float>(board, ch, "I0Set", val);
}

inline bool caenPST::setV1(
  unsigned int   board,
  unsigned short ch,
  float          val
) const {
  return setCVal<float>(board, ch, "V1Set", val);
}

inline bool caenPST::setI1(
  unsigned int   board,
  unsigned short ch,
  float          val
) const {
  return setCVal<float>(board, ch, "I1Set", val);
}

inline bool caenPST::setTrip(
  unsigned int   board,
  unsigned short ch,
  float          val
) const {
  return setCVal<float>(board, ch, "Trip", val);
}

inline bool caenPST::setRup(
  unsigned int   board,
  unsigned short ch,
  float          val
) const {
  return setCVal<float>(board, ch, "RUp", val);
}

inline bool caenPST::setRdn(
  unsigned int   board,
  unsigned short ch,
  float          val
) const {
  return setCVal<float>(board, ch, "RDWn", val);
}

inline bool caenPST::setSVMax(
  unsigned int   board,
  unsigned short ch,
  float          val
) const {
  return setCVal<float>(board, ch, "SVMax", val);
}

inline bool caenPST::setOnOff(
  unsigned int   board,
  unsigned short ch,
  bool           val
) const {
  return setCVal<int>(board, ch, "Pw", val);
}

inline bool caenPST::setOn(
  unsigned int   board,
  unsigned short ch
) const {
  return setCVal<int>(board, ch, "Pw", true);
}

inline bool caenPST::setOff(
  unsigned int   board,
  unsigned short ch
) const {
  return setCVal<int>(board, ch, "Pw", false);
}

#endif // SY2527_HH_
