//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// Test program for the SY2527 power supply
//
// Hardware setup:
// - Power:                     ON, LOCAL
// - "Local" switch:            LOCAL ENABLE
// - "Interlock" switch:        OFF
// - HV Board:                  50 Ohm dummy load on HV-ENABLE
// -------------------------------------------------------------------
// $Id: tester.cxx,v 1.9 2004/12/16 11:17:29 loreti Exp $
// -------------------------------------------------------------------

#include <cctype>
#include "sy2527.hh"

extern "C" {
  #include <unistd.h>
}

using std::cerr;
using std::cin;
using std::cout;
using std::endl;
using std::getline;
using std::isspace;
using std::string;
using std::strtoul;

namespace {
  const string       defaultName    = "Test";
  const string       defaultAddress = IPADDR;
  unsigned long      boardNr        = 1;
  unsigned int       channNr        = 0;
}

static void printBdStatus(
  unsigned int status
) {
  // Prints on the standard output the meaning of the bits of the
  // board status

  if (status & 0x01) cout << "FAIL"; else cout << "OK";
  if (status & 0x02) cout << ", CKSUMERROR";
  if (status & 0x04) cout << ", HVCALERROR";
  if (status & 0x08) cout << ", TEMPCALERR";
  if (status & 0x10) cout << ", UNDERTEMP";
  if (status & 0x20) cout << ", OVERTEMP";
  cout << endl;
}

static bool monitorBoard(
  const caenPST & ps,
  int             board
) {
  // Gets and prints board temperature and status; returns "true" if
  // all is OK (status == 0)

  double temp         = ps.getTemp(board);
  unsigned int status = ps.getBdStatus(board);

  cout << "Board " << board << ": Temp = " << temp
       << " degrees\nBoard status: ";
  printBdStatus(status);
  return status == 0;
}

static void printChStatus(
  unsigned int status
) {
  // Prints on the standard output the meaning of the bits of the
  // channel status

  if (status & 0x001) cout << "ON"; else cout << "OFF";
  if (status & 0x002) cout << ", RAMPUP";
  if (status & 0x004) cout << ", RAMPDOWN";
  if (status & 0x008) cout << ", OVERCURR";
  if (status & 0x010) cout << ", OVERVOLT";
  if (status & 0x020) cout << ", UNDERVOLT";
  if (status & 0x040) cout << ", EXTERNTRIP";
  if (status & 0x080) cout << ", MAXHV";
  if (status & 0x100) cout << ", EXTERNDISABLE";
  if (status & 0x200) cout << ", INTERNTRIP";
  if (status & 0x400) cout << ", CALERROR";
  if (status & 0x800) cout << ", UNPLUGGED";
  cout << endl;
}

static int monitorChannel(
  const caenPST & ps,
  int             board,
  int             channel
) {
  // Gets and prints relevant values and status; returns -1 in case of
  // error, 0 if all is OK and the channel is OFF, 1 if all is OK and
  // the channel is ON.

  double v            = ps.getV(board, channel);
  double i            = ps.getI(board, channel);
  double v0           = ps.getV0(board, channel);
  double i0           = ps.getI0(board, channel);
  double trip         = ps.getTrip(board, channel);
  double rup          = ps.getRup(board, channel);
  double rdn          = ps.getRdn(board, channel);
  double svmax        = ps.getSVMax(board, channel);
  unsigned int status = ps.getChStatus(board, channel);

  cout << "Bd " << board << ", ch " << channel
       << ": V=" << v << ", I=" << i << ", V0="
       << v0 << ", I0=" << i0 << ", Trip=" << trip
       << ", RUp=" << rup << ", Rdn=" << rdn
       << ", SVMax=" << svmax << "\nChannel status: ";
  printChStatus(status);
  if (status < 2) return status; else return -1;
}

int main() {
  string name, address, answer;

  cout << "Enter name (default \"" << defaultName << "\"): ";
  if (!getline(cin, name)) return 0;
  if (name.empty()) name = defaultName;

  cout << "Enter address (default \"" << defaultAddress << "\"): ";
  if (!getline(cin, address)) return 0;
  if (address.empty()) address = defaultAddress;

  while(true) {
    cout << "Enter board number (default " << boardNr << "): ";
    if (!getline(cin, answer)) return 0;
    if (answer.empty()) break;

    char *        pc;
    unsigned long b = strtoul(answer.c_str(), &pc, 0);

    if (!isspace(static_cast<unsigned char>(*pc)) && *pc != '\0') {
      cerr << "Invalid: \"" << answer << "\"\n";
    } else {
      boardNr = b;
      break;
    }
  }

  while(true) {
    cout << "Enter channel number (default " << channNr << "): ";
    if (!getline(cin, answer)) return 0;
    if (answer.empty()) break;

    char *        pc;
    unsigned long c = strtoul(answer.c_str(), &pc, 0);

    if (!isspace(static_cast<unsigned char>(*pc)) && *pc != '\0') {
      cerr << "Invalid: \"" << answer << "\"\n";
    } else {
      channNr = c;
      break;
    }
  }

  // - Prints the software version;
  // - Opens the connection toward the power supply;
  // - Prints the current settings for board "boardNr" and board
  //   "boardNr" channel "channNr"; that channel is assumed to be
  //   initially off.
  // - Drives board "boardNr" channel "channNr" output to 50V,
  //   monitoring the ramp-up;
  // - Drives board "boardNr" channel "channNr" off, monitoring the
  //   ramp-down.

  cout << '\n';
  sy2527::version();
  cout << endl;

  try {
    caenPST ps(name, address);

    if (!ps.testBC(boardNr, channNr)) {
      cerr << "\nBoard " << boardNr << " channel " << channNr
           << ": invalid\n";
      return 0;
    }

#ifndef NDEBUG
    cout << endl;
#endif
    monitorBoard(ps, boardNr);
    monitorChannel(ps, boardNr, channNr);

    ps.setV0(boardNr, channNr, 50.0);
    ps.setI0(boardNr, channNr, 10.0);
    ps.setTrip(boardNr, channNr, 10.0);
    ps.setRup(boardNr, channNr, 5.0);
    ps.setRdn(boardNr, channNr, 5.0);
    ps.setSVMax(boardNr, channNr, 60.0);
    ps.setOn(boardNr, channNr);
    sleep(1);
    for (;;) {
      if (monitorChannel(ps, boardNr, channNr) == 1) break;
      sleep(5);
    }
    cout << "\n*** BOARD " << boardNr << " CHANNEL "
         << channNr << " IS UP\n" << endl;

    sleep(5);

    ps.setOff(boardNr, channNr);
    sleep(1);
    for (;;) {
      if (monitorChannel(ps, boardNr, channNr) == 0) break;
      sleep(5);
    }
    cout << "\n*** BOARD " << boardNr << " CHANNEL "
         << channNr << " IS DOWN" << endl;

  } catch (const std::exception & e) {
    cerr << "*** EXCEPTION:" << e.what() << endl;
  }
}
